package weaving;

import org.eclipse.cdt.core.dom.ast.IASTDeclaration;
import org.eclipse.cdt.core.dom.ast.IASTFileLocation;
import org.eclipse.cdt.core.dom.ast.IASTName;
import org.eclipse.cdt.core.dom.ast.IASTNode;
import org.eclipse.cdt.core.model.CModelException;
import org.eclipse.cdt.core.model.ICElement;
import org.eclipse.cdt.internal.core.dom.parser.cpp.CPPASTFunctionDefinition;
import org.eclipse.cdt.internal.core.model.ASTStringUtil;
import org.eclipse.cdt.internal.core.model.CModelBuilder2;
import org.eclipse.cdt.internal.core.model.MethodDeclaration;
import org.eclipse.cdt.internal.core.model.Parent;
import org.eclipse.cdt.internal.core.model.Structure;

import parser.ACCAdvice;
import parser.ACCAspect;
import parser.ACCAttribute;
import parser.ACCPointcut;
import parser.ACCSlice;
import parser.IACCASTAdvice;
import parser.IACCASTAspect;
import parser.IACCASTAttribute;
import parser.IACCASTPointcut;
import parser.IACCASTSlice;


@SuppressWarnings("restriction")
aspect ModelBuilder {

	private boolean isSlice  = false;
	private boolean isAspect = false;
	
	Structure around(ICElement parent, int kind, String name) :
		call(Structure.new(..)) && args(parent, kind, name) {
		if (isSlice) {
			ACCSlice structure = new ACCSlice(parent, kind, name);
			return structure;
		}
		else if (isAspect) {
			ACCAspect structure = new ACCAspect(parent, kind, name);
			return structure;
		}
		else
			return proceed(parent, kind, name);
	}
	
	private void setBodyPosition(MethodDeclaration info, IASTNode astNode) {
		final IASTFileLocation location= astNode.getFileLocation();
		if (location != null) {
			info.setPos(location.getNodeOffset(), location.getNodeLength());
			info.setLines(location.getStartingLineNumber(), location.getEndingLineNumber());
		}
	}

	private void setIdentifierPosition(ACCAdvice info, IACCASTAdvice astName) {

		// set 'before', 'after', 'around', or 'before' as ID
		if (astName.getKey() != null &&
				astName.getKey().getFileLocation() != null) {
			IASTFileLocation location = astName.getKey().getFileLocation();
			info.setIdPos(location.getNodeOffset(), location.getNodeLength());
		}

	}

	void around(Parent parent, IASTDeclaration declaration) :
		execution (void CModelBuilder2.createDeclaration(..)) && args(parent, declaration) {
		if (declaration instanceof IACCASTAdvice) {
			IACCASTAdvice adviceDecl = (IACCASTAdvice)declaration;
			
			String functionName = "slice"; // default if no key is set
			if (adviceDecl.getKey() != null) {
				final IASTName name= adviceDecl.getKey();
		        if (name != null)
		        	functionName= ASTStringUtil.getSimpleName(name);
			}
			
			ACCAdvice adv = new ACCAdvice(parent, functionName);
			adv.setPointcut(adviceDecl.getPointcut());
			try {
				parent.addChild(adv);
				if (adviceDecl.getTemplateDeclaration() != null) {
					final String[] parameterTypes =
							ASTStringUtil.getParameterSignatureArray(((CPPASTFunctionDefinition)adviceDecl.getTemplateDeclaration().getDeclaration()).getDeclarator());
					adv.setParameterTypes(parameterTypes);
				}
				setBodyPosition(adv, declaration);
				setIdentifierPosition(adv, adviceDecl);
			} catch (CModelException e) {
				// TODO Auto-generated catch block
				System.out.println("GRRRR");
				e.printStackTrace();
			}
		}
		else if (declaration instanceof IACCASTPointcut) {
			IACCASTPointcut pointcutDecl = (IACCASTPointcut)declaration;
			
			String[] parameterTypes = null;
			String functionName = null;
			IASTName name = null;
			if (pointcutDecl.getFunctionDeclarator() != null) {
				parameterTypes =
						ASTStringUtil.getParameterSignatureArray(pointcutDecl.getFunctionDeclarator());
				name = pointcutDecl.getFunctionDeclarator().getName();
				functionName = ASTStringUtil.getSimpleName(name);
			}
			ACCPointcut pct = new ACCPointcut(parent, functionName);
			pct.setParameterTypes(parameterTypes);
			pct.setPointcut(pointcutDecl.getPointcut());
			try {
				parent.addChild(pct);
				setBodyPosition(pct, declaration);
				if (name != null && name.getFileLocation() != null) {
					IASTFileLocation location = name.getFileLocation();
					pct.setIdPos(location.getNodeOffset(), location.getNodeLength());
				}
			} catch (CModelException e) {
				// TODO Auto-generated catch block
				System.out.println("GRRRR");
				e.printStackTrace();
			}
		}
		else if (declaration instanceof IACCASTAttribute) {
			IACCASTAttribute attributeDecl = (IACCASTAttribute)declaration;
			
			String[] parameterTypes = null;
			String functionName = null;
			IASTName name = null;
			if (attributeDecl.getFunctionDeclarator() != null) {
				parameterTypes =
						ASTStringUtil.getParameterSignatureArray(attributeDecl.getFunctionDeclarator());
				name = attributeDecl.getFunctionDeclarator().getName();
				functionName = ASTStringUtil.getSimpleName(name);
			}
			ACCAttribute attr = new ACCAttribute(parent, functionName);
			attr.setParameterTypes(parameterTypes);
			try {
				parent.addChild(attr);
				setBodyPosition(attr, declaration);
				if (name != null && name.getFileLocation() != null) {
					IASTFileLocation location = name.getFileLocation();
					attr.setIdPos(location.getNodeOffset(), location.getNodeLength());
				}
			} catch (CModelException e) {
				// TODO Auto-generated catch block
				System.out.println("GRRRR");
				e.printStackTrace();
			}
		}
		else if (declaration instanceof IACCASTSlice) {
			IACCASTSlice sliceDecl = (IACCASTSlice)declaration;
			isSlice = true;
			proceed (parent, sliceDecl.getDeclaration());
			isSlice = false;
		}
		else if (declaration instanceof IACCASTAspect) {
			IACCASTAspect aspectDecl = (IACCASTAspect)declaration;
			isAspect = true;
			proceed (parent, aspectDecl.getDeclaration());
			isAspect = false;
		}
		else {
			boolean oldIsAspect = isAspect;
			isAspect = false;
			proceed (parent, declaration);
			isAspect = oldIsAspect;
		}
	}
	
}