package weaving;

import parser.ACCSourceParser;

import org.eclipse.cdt.core.dom.ast.IASTCompoundStatement;
import org.eclipse.cdt.core.dom.ast.IASTDeclaration;
import org.eclipse.cdt.core.parser.EndOfFileException;
import org.eclipse.cdt.internal.core.dom.parser.BacktrackException;
import org.eclipse.cdt.internal.core.parser.scanner.CPreprocessor;
import org.eclipse.cdt.internal.core.dom.parser.cpp.GNUCPPSourceParser;

privileged aspect ParserHelper {
	
	private int adviceNo;

    private void CPreprocessor.removeMacroDefinition(char[] key) {
    	fMacroDictionary.remove(key, 0, key.length);
    	fCurrentContext.internalModification(key);
    }
    
    before(): execution(* ACCSourceParser.aspectDeclaration(..)) {
    	adviceNo = 0;
    }
    
    IASTDeclaration around(): execution(* ACCSourceParser.adviceDeclaration(..)) {
		CPreprocessor cpp = (CPreprocessor)(((ACCSourceParser)thisJoinPoint.getThis()).scanner);
		
		String joinPointDef = "struct JoinPoint {" +
				"  typedef __T That;" +
				"  typedef __T Target;" +
				"  typedef __T Entity;" +
				"  typedef __T MemberPtr;" +
				"  typedef __T Result;" +
				"  struct Res {" +
				"    typedef __T Type;" +
				"    typedef __T ReferredType;" +
				"  };" +
				"  template <int I> struct Arg {" +
				"    typedef __T Type;" +
				"    typedef __T ReferredType;" +
				"  };" +
				"  enum { ARGS };" +
				"  typedef __T Array;" +
				"  template <int I> struct Dim {" +
				"    typedef __T Idx;" +
				"    enum { Size };" +
				"  };" +
				"  enum { DIMS };" +
				"  enum { JPID };" +
				"  enum { JPTYPE };" +
				"  static const char *signature();" +
				"  static const char *filename();" +
				"  static int line();" +
				"  That *that();" +
				"  Target *target();" +
				"  Entity *entity();" +
				"  MemberPtr memberptr();" +
				"  Result *result();" +
				"  template <int I> void *arg();" +
				"  void *arg(int i);" +
				"  void proceed();" +
				"  Array *array();" +
				"  template <int I> void *idx();" +
				"  static const char *resulttype();" +
				"  static const char *argtype(int i);" +
				"  struct Action {}; Action &action();" +
				"} * tjp;";

		String adviceFuncPrefix = "template <typename __T> void __a" +
				String.valueOf(adviceNo) + "_";
		cpp.addMacroDefinition("before(X)".toCharArray(),
				(adviceFuncPrefix+ "before(X) { " + joinPointDef).toCharArray());
		cpp.addMacroDefinition("after(X)".toCharArray(),
				(adviceFuncPrefix+ "after(X) { " + joinPointDef).toCharArray());
		cpp.addMacroDefinition("around(X)".toCharArray(),
				(adviceFuncPrefix+ "around(X) { " + joinPointDef).toCharArray());
		IASTDeclaration decl = proceed();
		cpp.removeMacroDefinition("before".toCharArray());
		cpp.removeMacroDefinition("after".toCharArray());
		cpp.removeMacroDefinition("around".toCharArray());
		adviceNo++;
		return decl;
    }
    
    IASTCompoundStatement around() throws EndOfFileException, BacktrackException : execution(* GNUCPPSourceParser.functionBody()) &&
    	cflow(execution(* ACCSourceParser.adviceDeclaration(..))) && 
    	!cflow(execution(* ACCSourceParser.adviceBody())) {
    	if (thisJoinPoint.getThis() instanceof ACCSourceParser) {
    		ACCSourceParser parser = (ACCSourceParser)thisJoinPoint.getThis();
    		return parser.adviceBody();
    	}
    	return proceed();
    }
}